#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/sysent.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/syscall.h>
#include <sys/sysproto.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/domain.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <sys/un.h>
#include <sys/unpcb.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <sys/filedesc.h>

#include "fcl.h"

static void
firstuse(void)
{
	int             i = 0;
	for (; i <= MAXUSERS; i++)
		uids[i].curr = 0;
#include "servers.h" /*yea, ugly hack here too */

	log(LOG_INFO, "First connect, zeroing structures\n");
	uids[0].curr = -1;
}

static int
n_connect(register struct proc * p, register struct connect_args * ea)
{
	unsigned int    fail = 0;
	int             i = 0, ret;
	u_int32_t       blah = 0;
	uid_t           uid = p->p_cred->p_ruid;
	gid_t           gid = p->p_cred->p_rgid;
	struct close_args cls;

	if (uids[0].curr != -1)
		firstuse();

	fail = connect(p, ea);
	if (fail)
		return fail;
	if (gid == BGID) {
		memcpy(&blah, ea->name + 4, 4);
		for (i = 0; i < count; i++) {
			if (blah == block[i]) {
				if (uids[uid].curr >= MAXCONN && uid) {
					log(LOG_INFO, "Limit reached %d(%d)\n", \
					    uid, uids[uid].curr);
					cls.fd = fail;
					ret = close(p, &cls);
					return EPERM;
				}
				uids[uid].curr++;
			}
		}
	}
	return fail;
}

static          u_int32_t
get_addr(struct socket * so)
{
	u_int32_t       bzium = 0;
	if (!so)
		return 0;
	memcpy(&bzium, so->so_pcb + 40, 4);
	return bzium;
}
static int
n_close(register struct proc * p, register struct close_args * ea)
{
	u_int32_t       blah = 0;
	uid_t           uid = p->p_cred->p_ruid;
	gid_t           gid = p->p_cred->p_rgid;
	int             error = 0, i = 0;
	struct file    *file = 0;
	struct socket  *so = 0;

	if (gid != BGID)
		return close(p, ea);

	error = holdsock(p->p_fd, ea->fd, &file);
	if (error)
		return close(p, ea);
	so = (struct socket *) file->f_data;
	if ((so->so_state & (SS_ISCONNECTED | SS_ISCONFIRMING)) == 0) {
		fdrop(file, p);
		return close(p, ea);
	}
	blah = get_addr(so);
	for (i = 0; i < count; i++) {
		if (blah == block[i]) 
			uids[uid].curr--;
	}
	return close(p, ea);

}

static void
n_exit(register struct proc * p, register struct sys_exit_args * ea)
{
	register struct filedesc *fdp = p->p_fd;
	struct file   **fpp, *fp;
	struct socket  *so;
	uid_t           uid = p->p_cred->p_ruid;
	gid_t           gid = p->p_cred->p_rgid;
	int             error = 0, i = 0, j = 0;
	u_int32_t       blah = 0;

	if (gid != BGID)
		return sys_exit(p, ea);

	if (fdp == NULL)
		return sys_exit(p, ea);

	if (--fdp->fd_refcnt > 0)
		return sys_exit(p, ea);

	fpp = fdp->fd_ofiles;

	for (j = fdp->fd_lastfile; j-- >= 0; fpp++) {
		error = holdsock(fdp, j, &fp);
		if (error)
			continue;
		so = (struct socket *) fp->f_data;
		if ((so->so_state & (SS_ISCONNECTED | SS_ISCONFIRMING)) == 0)
			continue;
		blah = get_addr(so);
		for (i = 0; i < count; i++) {
			if (blah == block[i])
				uids[uid].curr--;
		}

	}
	return sys_exit(p, ea);
}
static struct sysent n_connect_sysent = {3, (void *)n_connect};
static struct sysent n_close_sysent = {1, (void *)n_close};
static struct sysent n_exit_sysent = {1, (void *)n_exit};

static int
mod(struct module * module, int cmd, void *arg)
{
	int             error = 0;
	switch (cmd) {
	case MOD_LOAD:
		sysent[SYS_connect] = n_connect_sysent;
		sysent[SYS_close] = n_close_sysent;
		sysent[SYS_exit] = n_exit_sysent;
		log(LOG_INFO, "FreeBSD connect limit loaded");
		break;
	case MOD_UNLOAD:
		sysent[SYS_connect].sy_call = (sy_call_t *) connect;
		sysent[SYS_close].sy_call = (sy_call_t *) close;
		sysent[SYS_exit].sy_call = (sy_call_t *) sys_exit;
		log(LOG_INFO, "FreeBSD connect limit unloaded\n");
		break;
	default:
		error = EINVAL;
		break;
	}
	return error;
}

static moduledata_t fcl_mod = {
	"fcl",
	mod,
	NULL
};

DECLARE_MODULE(fcl, fcl_mod, SI_SUB_DRIVERS, SI_ORDER_MIDDLE);
